{-# OPTIONS -cpp #-}
{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Main where

import Prelude hiding		    ( print, putStr, putStrLn, catch )
import IO.Output

import Control.Exception	    ( catch, Exception(ErrorCall), evaluate, throwIO )
import Control.Monad
import Control.Monad.Error	    ( catchError, throwError )
import Control.Monad.Trans	    ( liftIO )
import Control.Monad.State
import qualified Data.Map as Map
import Language.Haskell.Pretty
import System			    ( exitFailure, getProgName, getArgs )
import Text.PrettyPrint

import Compilers.Haskell
import Compilers.Latex
import Compilers.Vim
import Compilers.Emacs
import Compilers.Run		    hiding (evaluate)

import Interface.Interactive
import Interface.Options            -- ( usage, ...)
import Interface.ChaseImports

import Plugins			    
import Plugins.All		    ()    -- To force compilation

import Syntax.ConcreteToInternal
import Syntax.ParseMonad
import Syntax.Parser
import qualified Syntax.Concrete as C

import TypeChecker		    ( checkProgram )
import TypeChecker.Monad
import TypeChecker.Monad.Options
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Statistics
import TypeChecker.Errors

import Utils.Error
import Utils.Monads
import Utils.Unicode
import Utils.UglyHack

concatPrograms :: [ParseResult C.Program] -> TC C.Program
concatPrograms rs =
    do	z <- liftIO $ (Right <$> evaluate (concatPrograms' rs))
		`catch` \e ->
		    case e of
			ErrorCall err	-> return (Left err)
			_		-> throwIO e
	case z of
	    Left err	-> fail err -- TODO: Better error
	    Right p	-> return p
    where
	concatPrograms' [] = []
	concatPrograms' (r:rs) =
	    case r of
		ParseFailed p why what -> error $ showParseFailure p why what
		ParseOk _ prog ->
		    prog ++ concatPrograms' rs

parseFiles :: TC C.Program
parseFiles =
    do	opts <- commandLineOptions
	rs <- liftIO $ mapM parseProgram (optInputFiles opts)
	concatPrograms rs
    where
	parseProgram file
	    | isLiterate file	= parseLiterateFile programParser file
	    | otherwise		= parseFile programParser file

runAgdaLight :: TC ()
runAgdaLight =
    do	-- Parse (standard) options. Plugin options are parsed
	-- when loading the plugins.
	progName    <- liftIO $ getProgName
	argv	    <- liftIO $ getArgs
	let opts' = parseStandardOptions
			progName
			[ a | a <- argv, notPluginOption a ]
	case opts' of
	    Left err	-> liftIO $ optionError err
	    Right opts
		| optShowHelp opts	-> liftIO printUsage
		| optShowVersion opts	-> liftIO printVersion
		| null (optInputFiles opts)
					-> liftIO printUsage
		| length (optInputFiles opts) > 1
					-> liftIO multipleFilesDeprecated
		| optTexify opts	-> liftIO $ toLaTeX opts
		| otherwise		-> checkFiles opts
    where
	checkFiles opts = do
		liftIO $ setUglyHack "show-implicit"
		       $ if optShowImplicit opts then 1 else 0
		files <- liftIO $ chaseImports	(optIncludeDirs opts)
						(optInputFiles opts)
		when (optVerbose opts >= 1) $ liftIO $
		    do	putStrLn "After import chasing:"
			mapM_ (\file -> putStrLn $ "  " ++ file) files
		setCommandLineOptions (opts { optInputFiles = files }) $
		    do  when (optInteractive opts) $
			    liftIO $ putStr splashScreen
		    

			(if optInteractive opts
				then interactionLoop
				else \m -> m >> return () ) $
			    do	-- Clear the state
				clearState

				-- Load plugins
				loadPlugins `catchError` \e ->
				    case errCode e of
					OptionError -> liftIO $ optionError
								    (show $ errWhy e)
					_	    -> throwError e

				-- Get options
				opts <- commandLineOptions
			    
				-- Parse input files
				prog <- parseFiles

				-- Convert to internal syntax
				(iprog, (used, scope)) <-
				    case programToInternal prog of
					Left err    -> sayWhen whenCheckingScope $ fail err
					Right x	    -> return x

				withScopeEnv scope $
				    do	-- Update used names
					setUsedNames used

					-- Generate name files
					when (optGenVimFile opts) $
					    liftIO $ generateVimFile (last $ optInputFiles opts) iprog
					when (optGenEmacsFile opts) $
					    liftIO $ generateEmacsFile (last $ optInputFiles opts) iprog

					-- Print parsed program
					when (optVerbose opts >= 5) $
					    liftIO $ print iprog

					-- Type check
					defs <- checkProgram iprog

					-- Let plugins finish up
					finalizePlugins

					-- Print statistics
					verbose 2 $
					    do  stats <- getStatistics
						let pr (s,n) = putStrLn $ "  " ++ s ++ ": " ++ show n
						liftIO $
						    case Map.assocs stats of
							[] -> return ()
							xs ->
							    do putStrLn "Statistics:"
							       mapM_ pr xs
					-- Check that we don't have any left over constraints
					cs <- gets tcPostponedConstraints
					case cs of
					    []	-> return ()
					    _	-> failure $ customError $ show (length cs) ++ " left over constraints!"

					return scope
			when (optCompile opts) $
			    liftIO $ putStrLn "Compilation to Haskell is currently broken. Sorry for the inconvenience."
-- 			    do	m <- compile
-- 				let file    = last $ optInputFiles opts
-- 				    outfile = maybe (file ++ ".hs") id
-- 						    (optOutputFile opts)
-- 				liftIO $ writeFile outfile $ prettyPrint m
			return ()

optionError :: String -> IO ()
optionError err =
    do	putStr $ "Bad arguments: " ++ err
	printUsage
	exitFailure

multipleFilesDeprecated :: IO ()
multipleFilesDeprecated =
    do	print $
	    text "NOTE:" $$ nest 2 (vcat
		[ fsep $ pwords $
		    "The passing of multiple files to agdaLight is " ++
		    "no longer supported. To achieve the same effect " ++
		    "use the new import pragmas:"
		, nest 2 $ text "-- # import Property"
		, nest 2 $ text "-- # import QuickCheck"
		, nest 2 $ text "-- # import examples.fol.List"
		, fsep $ pwords $
		    "Imports are looked for in"
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "the same directory as the file " ++
					    "containing the import,"
				)
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "the current working directory,")
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "any directory specified using the -i flag")
		, nest 2 $ text "-" <+>
			    fsep (pwords $ "the agdaLight lib directory (" ++
					    _ROOT_ ++ "/lib) directory.")
		, fsep $ pwords $ "To find a file the dots are replaced by slashes and '.agda'" ++
				    " is appended to the module name, so examples.fol.List" ++
				    " becomes examples/fol/List.agda."
		]
	    )
	exitFailure

toLaTeX :: Options -> IO ()
toLaTeX opts =
    do	files <- liftIO $ chaseImports	(optIncludeDirs opts)
					(optInputFiles opts)
	let file    = last $ files
	    outfile = maybe (file ++ ".tex") id
			    (optOutputFile opts)
	writeFileUTF8 outfile =<< texify files

printVersion :: IO ()
printVersion =
    do	putStrLn $ "AgdaLight version " ++ _VERSION_
	putStrLn "  Andreas Abel, Thierry Coquand, Ulf Norell"

printUsage :: IO ()
printUsage =
    do	progName <- getProgName
	putStr $ usage standardOptions_ allPluginInfo progName

main :: IO ()
main =
    do	r <- runTC runAgdaLight
	case r of
	    Left err	-> do	print err
				exitFailure
	    Right ()	-> return ()

